package net.sf.sido.pojo;

import net.sf.sido.DataFactory;
import net.sf.sido.factory.sido.SiDOLDataFactoryReader;
import org.apache.commons.lang.reflect.ConstructorUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.*;

public class DOFactory {

    public static final class DefaultDOFactoryLogger implements DOFactoryLogger {

        private final Logger logger = LoggerFactory.getLogger(DOFactory.class);

        @Override
        public void log(String pattern, Object... parameters) {
            logger.info(String.format(pattern, parameters));
        }
    }

    public static final String SIDO_POJO_RESOURCE_PATH = "META-INF/sido/sido.pojo";
    public static final String SIDOL_RESOURCE_PATH = "META-INF/sido/%s.sidol";

    private final DOFactoryLogger logger;
    private final DataFactory dataFactory;
    private final Map<Class<? extends DOSchema>, DOSchema> schemas = new HashMap<Class<? extends DOSchema>, DOSchema>();

    public DOFactory(DataFactory dataFactory) {
        this(dataFactory, false);
    }

    public DOFactory(DataFactory dataFactory, boolean discover) {
        this(dataFactory, discover, new DefaultDOFactoryLogger());
    }

    public DOFactory(DataFactory dataFactory, boolean discover, DOFactoryLogger logger) {
        this.logger = logger;
        this.dataFactory = dataFactory;
        // Gets all schemas
        if (discover) {
            try {
                discoverSchemas();
            } catch (IOException e) {
                throw new IllegalStateException("Cannot discover schemas", e);
            }
        }
    }

    public Collection<DOSchema> getSchemas() {
        return schemas.values();
    }

    protected ClassLoader getClassLoader() {
        ClassLoader cl = Thread.currentThread().getContextClassLoader();
        return cl != null ? cl : getClass().getClassLoader();
    }

    protected void discoverSchemas() throws IOException {
        logger.log("Discovering all schemas...");
        Map<String, URL> schemasURLs = new HashMap<String, URL>();
        SiDOLDataFactoryReader reader = new SiDOLDataFactoryReader();
        // Gets all sido.pojo resources
        Enumeration<URL> sidoPojoURLs = getClassLoader().getResources(SIDO_POJO_RESOURCE_PATH);
        while (sidoPojoURLs.hasMoreElements()) {
            URL sidoPojoURL = sidoPojoURLs.nextElement();
            logger.log("Loading POJO schema definitions from %s", sidoPojoURL);
            InputStream sidoPojoStream = sidoPojoURL.openStream();
            try {
                // Reads as properties
                Properties sidoPojoProperties = new Properties();
                sidoPojoProperties.load(sidoPojoStream);
                // For all properties
                for (String schemaName : sidoPojoProperties.stringPropertyNames()) {
                    String doSchemaClassName = sidoPojoProperties.getProperty(schemaName);
                    logger.log("Loading schema [%s] foo DOSchema [%s]", schemaName, doSchemaClassName);
                    // Loads the corresponding SiDOL file
                    String sidolResourcePath = String.format(SIDOL_RESOURCE_PATH, schemaName);
                    URL sidolResourceURL = getClassLoader().getResource(sidolResourcePath);
                    if (sidolResourceURL == null) {
                        throw new IllegalStateException(String.format("Cannot find resource for schema %s at %s", schemaName, sidolResourcePath));
                    }
                    // Registers the resource
                    schemasURLs.put(doSchemaClassName, sidolResourceURL);
                }
            } finally {
                sidoPojoStream.close();
            }
        }
        // Loads all URLs first
        reader.read(schemasURLs.values(), dataFactory);
        // Loads all DOSchemas
        for (String doSchemaClassName : schemasURLs.keySet()) {
            try {
                // Creates the schema
                Class<? extends DOSchema> doSchemaClass = (Class<? extends DOSchema>) Class.forName(doSchemaClassName, true, getClassLoader());
                DOSchema doSchema = (DOSchema) ConstructorUtils.invokeConstructor(doSchemaClass, this);
                // Makes sure it is registered
                register(doSchema);
            } catch (Exception ex) {
                throw new RuntimeException(String.format("Cannot load DOSchema %s", doSchemaClassName), ex);
            }
        }
    }

    public DataFactory getDataFactory() {
        return dataFactory;
    }

    public <S extends DOSchema> S getSchema(Class<S> schemaClass) {
        DOSchema doSchema = schemas.get(schemaClass);
        if (doSchema != null) {
            return (S) doSchema;
        } else {
            throw new IllegalStateException(String.format("Cannot find DOSchema of class %s. Maybe it has not been instantiated in the DOFactory.", schemaClass.getName()));
        }
    }

    void register(DOSchema doSchema) {
        schemas.put(doSchema.getClass(), doSchema);
    }
}
